<?php

namespace Iwaydev\BaseSender;

use DomainException;
use GuzzleHttp\Client;
use GuzzleHttp\RequestOptions;
use JsonException;
use Psr\Http\Message\ResponseInterface;
use Psr\Log\LoggerInterface;
use Psr\SimpleCache\CacheInterface;

/**
 * Сендер для отправки запросов
 *
 * Class TransNextGenSender
 *
 * @package App\Senders
 */
class Sender
{
    public const BODY_FORMAT_JSON = 'json';
    public const BODY_FORMAT_TEXT = 'text';

    /** @var bool Возвращать ли ошибки в ответе */
    private bool $httpErrors = false;

    private Client $client;
    private string $version;

    private bool $caching = false;
    private bool $verify = true;
    private ?CacheInterface $cache;
    private int $cacheLifetime;

    private ?LoggerInterface $logger;
    private ?LoggerInterface $defaultLogger;

    protected array $headers = [];
    private string $xUuid;
    //Формат тела запроса, по-умолчанию json
    protected string $bodyFormat = self::BODY_FORMAT_JSON;

    /**
     * Sender constructor.
     *
     * @param string $url URI запроса
     * @param string $version Версия используемого API
     * @param CacheInterface|null $cache Инстанс кэшера
     * @param int $cacheLifetime Время жизни кэша в секундах
     * @param LoggerInterface|null $logger Инстанс логгера
     * @param string|null $xUuid UUID
     */
    public function __construct(
        string $url,
        string $version,
        ?CacheInterface $cache = null,
        int $cacheLifetime = 3600,
        ?LoggerInterface $logger = null,
        ?string $xUuid = null
    ) {
        $this->client = new Client([
            'base_uri' => $url,
            'http_errors' => false
        ]);
        $this->version = trim($version, '/');

        $this->cache = $cache;
        $this->cacheLifetime = $cacheLifetime;

        $this->setLogger($logger);
        $this->setXUuid($xUuid);
    }

    /**
     * Возвращает текущий http-клиент
     *
     * @return Client
     */
    public function getClient(): Client
    {
        return $this->client;
    }

    /**
     * Возвращать ошибки из ответа
     *
     * @return $this
     */
    public function withErrors(): self
    {
        $this->httpErrors = true;

        return $this;
    }

    /**
     * Игнорить SSL
     *
     * @return $this
     */
    public function withoutVerify(): self
    {
        $this->verify = false;

        return $this;
    }

    /**
     * Отправить GET-запрос
     *
     * @param string $uri URI запроса
     * @param array $query GET-параметры
     *
     * @return array
     * @throws JsonException
     */
    protected function get(string $uri, array $query = []): array
    {
        return $this->send('GET', $uri, ['query' => $query]);
    }

    /**
     * Отправить POST-запрос
     *
     * @param string $uri URI запроса
     * @param array $body Данные запроса
     *
     * @return array
     * @throws JsonException
     */
    protected function post(string $uri, array $body = []): array
    {
        return $this->send('POST', $uri, ['json' => $body]);
    }

    /**
     * Отправить PATCH-запрос
     *
     * @param string $uri URI запроса
     * @param array $body Данные запроса
     *
     * @return array
     * @throws JsonException
     */
    protected function patch(string $uri, array $body = []): array
    {
        return $this->send('PATCH', $uri, ['json' => $body]);
    }

    /**
     * Отправить PUT-запрос
     *
     * @param string $uri URI запроса
     * @param array $body Данные запроса
     *
     * @return array
     * @throws JsonException
     */
    protected function put(string $uri, array $body = []): array
    {
        return $this->send('PUT', $uri, ['json' => $body]);
    }

    /**
     * Отправить DELETE-запрос
     *
     * @param string $uri URI запроса
     * @param array $body Данные запроса
     *
     * @return array
     * @throws JsonException
     */
    protected function delete(string $uri, array $body = []): array
    {
        return $this->send('DELETE', $uri, ['json' => $body]);
    }

    /**
     * Кэшировать следующий запрос
     *
     * @return \App\Senders\Sender
     */
    public function caching(): self
    {
        $this->caching = true;

        return $this;
    }

    /**
     * Устанвить формат тела запроса текст
     *
     * @return \App\Senders\Sender
     */
    public function setBodyFormatText(): self
    {
        $this->bodyFormat = self::BODY_FORMAT_TEXT;

        return $this;
    }

    /**
     * Установить формат тела запроса json
     *
     * @return \App\Senders\Sender
     */
    public function setBodyFormatJson(): self
    {
        $this->bodyFormat = self::BODY_FORMAT_JSON;

        return $this;
    }

    /**
     * Отправка запроса
     *
     * @param string $method Тип запроса GET, POST и пр.
     * @param string $uri URI запроса
     * @param array $options Параметры запроса
     *
     * @return array
     * @throws JsonException
     */
    protected function send(string $method, string $uri, array $options = []): array
    {
        try {
            $options = array_merge($options, ['headers' => $this->headers]);
            $uri = $this->prepareUri($uri);
            $fullUrl = $this->client->getConfig('base_uri') . $uri;

            $this->logInfo('Request. Method: ' . $method);
            $this->logInfo('Request. Url: ' . $fullUrl);
            $this->logInfo('Request. Options:', $options);

            $key = (string)$fullUrl . json_encode($options);

            if (!empty($this->cache) && $this->caching === true) {
                $result = $this->cache->get($key);
                if (!empty($result)) {
                    $this->logInfo('Response. Is caching!');
                    $this->logInfo('Response. Content:', $result);

                    return $result;
                }
            }
            if ($this->httpErrors) {
                $options[RequestOptions::HTTP_ERRORS] = true;
                $this->httpErrors = false;
            }
            if ($this->verify === false) {
                $options[RequestOptions::VERIFY] = false;
                $this->verify = true;
            }

            $response = $this->client->request($method, $uri, $options);
            $result = $this->prepareResponse($response);

            if (!empty($this->cache) && $this->caching === true) {
                $this->cache->set($key, $result, $this->cacheLifetime);
                $this->caching = false;
            }

            $this->logInfo('Response. Content:', $result);

            $this->setLogger($this->defaultLogger);

            return $result;
        } catch (\Throwable $throwable) {
            $this->logException($throwable);

            throw $throwable;
        }
    }

    /**
     * Подготовка uri
     *
     * @param string $uri URI запроса
     *
     * @return string
     */
    protected function prepareUri(string $uri): string
    {
        return $uri ? DIRECTORY_SEPARATOR . $this->prepareVersion() . trim($uri, '/') : '';
    }

    /**
     * Возвращает версию апи, подготовленную для подстановки в uri
     *
     * @return string
     */
    private function prepareVersion(): string
    {
        if (empty($this->version)) {
            return '';
        }

        return $this->version . DIRECTORY_SEPARATOR;
    }

    /**
     * Подготовка ответа
     *
     * @param ResponseInterface $response Инстанс ответа
     *
     * @return array
     * @throws JsonException
     */
    protected function prepareResponse(ResponseInterface $response): array
    {
        if ($this->bodyFormat === self::BODY_FORMAT_JSON) {
            return json_decode(
                $response->getBody(),
                true,
                512,
                JSON_THROW_ON_ERROR
            );
        } else {
            //Возвращаем формат тела запроса json (по-умолчанию)
            $this->setBodyFormatJson();

            return [
                'body' => $response->getBody()->getContents()
            ];
        }
    }

    /**
     * Устанавливает логгер
     *
     * @param LoggerInterface|null $logger Инстанс логгера
     *
     * @return $this
     */
    public function setLogger(?LoggerInterface $logger): self
    {
        $this->logger = $logger;

        if (empty($this->defaultLogger)) {
            $this->defaultLogger = $logger;
        }

        return $this;
    }

    /**
     * Сделать запись уровня "info" в лог, если логгер установлен
     *
     * @param string $message Сообщение
     * @param array $context Дополнительные данные
     */
    private function logInfo(string $message, array $context = []): void
    {
        if (!empty($this->logger)) {
            $this->logger->info($this->xUuid . ' ' . $message, $context);
        }
    }

    /**
     * Сделать запись уровня "error" в лог, если логгер установлен
     *
     * @param string $message Сообщение
     * @param array $context Дополнительные данные
     */
    private function logError(string $message, array $context = []): void
    {
        if (!empty($this->logger)) {
            $this->logger->error($this->xUuid . ' ' . $message, $context);
        }
    }

    /**
     * Логируем исключение
     *
     * @param \Throwable $throwable Экзепляр исключения
     */
    public function logException(\Throwable $throwable): void
    {
        $this->logError('');
        $this->logError('Исключение!');
        $this->logError('Сообщение: ' . $throwable->getMessage());
        $this->logError('Код: ' . $throwable->getCode());
        $this->logError('Файл: ' . $throwable->getFile() . '(' . $throwable->getLine() . ')');
        $this->logError('Трейс: ' . PHP_EOL . $throwable->getTraceAsString());
        $this->logError('');
    }

    /**
     * Устанавливает заголовок
     *
     * @param string $header Заголовок
     * @param string $value Значение
     *
     * @return $this
     */
    public function setHeader(string $header, string $value): self
    {
        $this->headers[$header] = $value;

        return $this;
    }

    /**
     * Возвращает заголовок, если он есть
     *
     * @param string $name
     *
     * @return string
     */
    public function getHeader(string $name): string
    {
        if (isset($this->headers[$name])) {
            return $this->headers[$name];
        }

        throw new DomainException('Header ' . $name . ' not found');
    }

    /**
     * Устанавливает либо генерирует заголовок X-UUID
     *
     * @param string|null $xUuid Значение которое будет установлено в качестве заголовка X-UUID
     *
     * @return $this
     */
    public function setXUuid(?string $xUuid = null): self
    {
        $this->xUuid = empty($xUuid) ? md5(microtime()) : $xUuid;

        $this->setHeader('X-UUID', $this->xUuid);

        return $this;
    }

    /**
     * Возвращает текущий заголовок X-UUID
     *
     * @return string
     */
    public function getXUuid(): string
    {
        return $this->getHeader('X-UUID');
    }
}
